#  SPDX-FileCopyrightText: © 2022 Josef Hahn
#  SPDX-License-Identifier: AGPL-3.0-only
import json
import os
import subprocess
import sys
import typing as t

import hallyd.fs as _fs
import hallyd.lang as _lang
import hallyd.bindle as _bindle
import hallyd.typing as _typing


def check_call_with_stdin_string(cmd: list, *, stdin: t.AnyStr, **kwargs) -> None:
    stdin = stdin.encode() if isinstance(stdin, str) else stdin
    p = subprocess.Popen(cmd, stdin=subprocess.PIPE, **kwargs)
    stdout_result, stderr_result = p.communicate(stdin)
    if p.returncode:
        raise subprocess.CalledProcessError(p.returncode, cmd, stdout_result, stderr_result)
    return stdout_result


def check_output_with_stdin_string(cmd: list, *, stdin: t.AnyStr, **kwargs) -> bytes:
    return check_call_with_stdin_string(cmd, stdin=stdin, stdout=subprocess.PIPE, **kwargs)


def start_function_in_new_process(func: _typing.CallableWithQualifiedName, args=None, kwargs=None, *,
                                  interactive=False, capture_output=False, shell=()) -> subprocess.Popen:
    return subprocess.Popen(
        [*shell, sys.executable, _fs.Path(__file__).parent("_aux/subprocess_helper__call_function.py"),
         _lang.unique_id()],
        start_new_session=True,
        env={
            **os.environb,
            b"HALLYD_SUBPROCESS_DATA": _bindle.dumps({
                "MODULE_NAME": func.__module__,
                "FUNCTION_NAME": func.__qualname__,
                "ARGS": args or (),
                "KWARGS": kwargs or {},
            }),
            b"HALLYD_SUBPROCESS_SYS_PATH": json.dumps(list(sys.path))
        },
        stdin=None if interactive else subprocess.DEVNULL,
        stderr=None if interactive else subprocess.STDOUT,
        stdout=None if interactive else (subprocess.PIPE if capture_output else subprocess.DEVNULL))


def process_permanent_id_for_pid(pid: int) -> str:
    try:
        cmdline = _fs.Path(f"/proc/{pid}/cmdline").read_text()
        return _bindle.dumps({"PID": pid, "CMDLINE": cmdline})
    except FileNotFoundError:
        if _fs.Path("/proc/version").exists():
            raise _ProcessNotRunningError() from None
        raise _ProcessInfoUnavailableError() from None


def pid_for_process_permanent_id(process_permanent_id: str) -> int:
    return _bindle.loads(process_permanent_id)["PID"]


def is_process_running(process_permanent_id: str) -> t.Optional[bool]:
    pid = pid_for_process_permanent_id(process_permanent_id)
    try:
        return process_permanent_id_for_pid(pid) == process_permanent_id
    except _ProcessInfoUnavailableError:
        return None
    except _ProcessNotRunningError:
        return False


def verify_tool_available(tool: str) -> None:  # TODO use more
    """
    Checks if a tool is installed (by calling it). Raises an exception if not.
    """
    try:
        subprocess.check_output([tool, "--help"], stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        pass
    except IOError:
        raise RuntimeError(f"'{tool}' is not available") from None


class _ProcessNotRunningError(Exception):
    pass


class _ProcessInfoUnavailableError(Exception):
    pass
